/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <string.h>
#include <ctype.h>
#include "pan.h"

#include "xpm/book_open.xpm"
#include "xpm/book_closed.xpm"
#include "xpm/mini_page.xpm"
#include "xpm/greencheck.xpm"
#include "xpm/bluecheck.xpm"
#include "xpm/x.xpm"
#include "xpm/binary.xpm"
#include "xpm/binary_complete.xpm"
#include "xpm/binary_incomplete.xpm"

/**********************************************************
 * Private Variables
 **********************************************************/

static GnomePixmap *articlelist_closed_pixmap = NULL;
static GnomePixmap *articlelist_open_pixmap = NULL;
static GnomePixmap *decode_queue_pixmap = NULL;
static GnomePixmap *download_queue_pixmap = NULL;
static GnomePixmap *mp_complete_pixmap = NULL;
static GnomePixmap *mp_incomplete_pixmap = NULL;
static GnomePixmap *error_pixmap = NULL;
static GnomePixmap *article_read_pixmap = NULL;
static GnomePixmap *binary_decoded_pixmap = NULL;

GtkStyle *unread_style = NULL;
GtkStyle *read_style = NULL;
GtkStyle *new_style = NULL;

static GtkWidget *find_dialog;

/**********************************************************
 * Public Variables
 **********************************************************/

GdkColormap *cmap;
GdkColor new_color;
GdkColor read_color;
GdkColor unread_color;

article_data *current_article = NULL;
GtkWidget *article_ctree_menu;

/**********************************************************
 * Private Functions
 **********************************************************/

/* sorting the ui tree */
static gint articlelist_compare_subjects ( GtkCList*, gconstpointer, gconstpointer );
static gint articlelist_compare_text (GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2);
static gint articlelist_compare_lines (GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2); 

/* threading */
static int mp_strcasecmp ( const char *mp, const char *ph );
static const char* skip_reply_leader ( const char* a );
static int article_subject_comp ( const void* va, const void* vb );
static int gslist_message_id_comp ( const void* va, const void* vb );
static int gslist_subject_comp ( const void* va, const void* vb );
static int gslist_multipart_comp ( const void* va, const void* vb );

/* ui */
static GnomePixmap* get_article_pixmap ( const article_data* a );
static void articlelist_button_press (GtkWidget* widget, GdkEventButton* bevent);
static void articlelist_click_column (GtkCList *, int); /* int = which column was clicked */
static void articlelist_process_selected (int type);
static void article_ctree_destroy_cb ( void );
static void articlelist_sort ( void );
static void articlelist_sort_init ( group_data *gdata );
static void articlelist_init_styles (void);


/**********************************************************
 * Articlelist popup menu
 **********************************************************/

static GnomeUIInfo articlelist_menu_popup[] = {
	{ GNOME_APP_UI_ITEM, N_("Save Attachment"), N_("Save Attachments."), articlelist_selected_decode,
	  NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 'D', GDK_CONTROL_MASK, NULL },
	{ GNOME_APP_UI_ITEM, N_("Open Attachment"), N_("Open Attachments."), articlelist_selected_open,
	  NULL, NULL, GNOME_APP_PIXMAP_NONE, NULL, 'O', GDK_CONTROL_MASK, NULL },
	{ GNOME_APP_UI_ITEM, N_("Delete Message"), N_("Delete message from database."), article_delete },
	GNOMEUIINFO_SEPARATOR,
	{ GNOME_APP_UI_ITEM, N_("Mark as Read"), N_("Marks current message as read."), articlelist_selected_mark_read },
	{ GNOME_APP_UI_ITEM, N_("Mark as Unread"), N_("Marks the current message as unread."), articlelist_selected_mark_unread },
	GNOMEUIINFO_END
};

static GtkCTreeNode* find_article_next_pos = NULL;


static
void
articlelist_button_press (GtkWidget* widget, GdkEventButton* bevent)
{
	int row;
	int column;
	article_data *adata;

	if (bevent->button != 1)
		gtk_signal_emit_stop_by_name (GTK_OBJECT(widget),
					      "button_press_event");

	if (bevent->button == 1) {
		/* if in paned mode and one_click_preview set, 
		 * a single click shows message 
		 */
		if (one_click_preview && (viewmode == GUI_PANED)) {
			gtk_clist_get_selection_info (GTK_CLIST(widget), bevent->x,
						      bevent->y, &row, &column);
			adata = (article_data *) gtk_clist_get_row_data (GTK_CLIST(widget), row);
			if (adata) {
				current_article = adata;
				if (adata->body == NULL) {
					queue_add_article_download (adata,
								    QUEUE_ARTICLE_DOWNLOAD |
								    QUEUE_FOREGROUND);
					adata->state |= STATE_DOWNLOAD_QUEUED;
//					articlelist_update_node (adata);
				}
				else {
					text_update (adata);
				}
			}
		}
		
		/* notebook layout uses double clicking */
		if (bevent->type == GDK_2BUTTON_PRESS) {
			gtk_signal_emit_stop_by_name (GTK_OBJECT(widget),
						      "button_press_event");
			gtk_clist_get_selection_info (GTK_CLIST(widget), bevent->x,
						      bevent->y, &row, &column);
			adata = (article_data *) gtk_clist_get_row_data (GTK_CLIST(widget), row);
			if (adata) {
				current_article = adata;
				if (adata->body == NULL) {
					queue_add_article_download (adata,
								    QUEUE_ARTICLE_DOWNLOAD |
								    QUEUE_FOREGROUND);
					adata->state |= STATE_DOWNLOAD_QUEUED;
//					articlelist_update_node (adata);
				}
				else {
					text_update (adata);
				}
			}
			/* if in notebook layout, flip to the message tab */
			gui_page_set (MESSAGE_PAGE, text);
		}
	}

	else if (bevent->button == 3) {
		const gboolean sel = GTK_CLIST (Pan.article_ctree)->selection != NULL;

		GList* l = GTK_MENU_SHELL(article_ctree_menu)->children; /* l is "save attachment" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel );
		l = l->next; /* l is "open attachment" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel );
		l = l->next; /* l is "delete message" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel );
		l = l->next; /* l is separator */
		l = l->next; /* l is "mark as read" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel );
		l = l->next; /* l is "mark as unread" */
		gtk_widget_set_sensitive ( GTK_WIDGET(l->data), sel );

		gtk_menu_popup (GTK_MENU (article_ctree_menu), NULL, NULL, NULL,
				NULL, bevent->button, bevent->time);
	}
}


/**********************************************************
 * The following stuff implements the "find" functionality
 **********************************************************/

void
find_dialog_destroy_cb (void)
{
	gtk_widget_destroy(find_dialog);
	find_dialog = NULL;
}




typedef struct
{
	const gchar* author;
	const gchar* subject;
	gboolean case_sensitive;
}
header_compare_data;

static gint
find_compare_func (
	gconstpointer a,
	gconstpointer b )
{
	const article_data* adata = (const article_data*)a;
	const header_compare_data* header = (const header_compare_data*)b;

	/* if it ain't an article, it ain't a match */
	if ( !adata )
		return -1;

	/* make sure the author (if user specified one) matches.. */
	if ( header->author ) {
		if ( !adata->author )
			return -1;
		if ( header->case_sensitive ) {
			if ( !strstr(adata->author,header->author) )
				return -1;
		} else {
			if ( !stristr(adata->author,header->author) )
				return -1;
		}
	}

	/* make sure the subject (if user specified one) matches.. */
	if ( header->subject ) {
		if ( !adata->subject )
			return -1;
		if ( header->case_sensitive ) {
			if ( !strstr(adata->subject,header->subject) )
				return -1;
		} else {
			if ( !stristr(adata->subject,header->subject) )
				return -1;
		}
	}

	/* hooray, a match */
	return 0;
}



void
find_article (
	const gchar* author,
	const gchar* subject,
	gboolean case_sensitive )
{
	GtkCTreeNode *node, *n;

	header_compare_data header_data;
	header_data.author = author;
	header_data.subject = subject;
	header_data.case_sensitive = case_sensitive;

	node = gtk_ctree_find_by_row_data_custom (
		GTK_CTREE(Pan.article_ctree),
		find_article_next_pos,
		&header_data, 
		(GCompareFunc)find_compare_func );


	if ( !node ) {
		find_article_next_pos = NULL;
		return;
	}

	/* select the node.. */
	gtk_clist_freeze (GTK_CLIST(Pan.article_ctree));
	gtk_ctree_unselect_recursive (GTK_CTREE(Pan.article_ctree), NULL);
	gtk_ctree_select (GTK_CTREE(Pan.article_ctree), node);

	/* Make sure selected node is visible */
	for (n = GTK_CTREE_ROW(node)->parent; n; n = GTK_CTREE_ROW(n)->parent)
		gtk_ctree_expand(GTK_CTREE(Pan.article_ctree), n);
	gtk_ctree_node_moveto (GTK_CTREE(Pan.article_ctree), node, 0, 0, 0);
	gtk_clist_thaw (GTK_CLIST(Pan.article_ctree));
		     
	/* Find next node, paying attention to recursion */
	if (GTK_CTREE_ROW(node)->children) {
		find_article_next_pos = GTK_CTREE_ROW(node)->children;
	} else {
		for (n = node; n; n = GTK_CTREE_ROW(n)->parent) {
			if (GTK_CTREE_ROW(n)->sibling) {
				find_article_next_pos = GTK_CTREE_ROW(n)->sibling;
				break;
			}
		}
	}
}

static gchar* find_article__author = NULL;
static gchar* find_article__subject = NULL;
static gboolean find_article__case_sensitive = FALSE;

void
find_dialog_clicked_cb (GnomeDialog *dialog, int button, GtkWidget* table )
{
 	g_return_if_fail ( button==0 || button==1 );
	g_return_if_fail ( table!=NULL );

	if ( button==1 ) /* close button */
	{
		find_dialog_destroy_cb ( );
	}
	else if ( button==0 ) /* "find next" button */
	{
		GtkWidget* author_entry;
		GtkWidget* subject_entry;
		GtkWidget* case_sensitive_checkbox;
		gchar* author = 0;
		gchar* subject = 0;
		gboolean case_sensitive;

		/* get author to search for, or NULL if none... */
		author_entry = (GtkWidget*) gtk_object_get_data ( GTK_OBJECT(find_dialog), "author" );
		g_return_if_fail ( author_entry!=NULL );
		author = g_strdup ( gtk_entry_get_text ( GTK_ENTRY(author_entry) ) );
		g_strstrip ( author );
		if ( !*author ) {
			g_free ( author );
			author = NULL;
		}

		/* get subject to search for, or NULL if none... */
		subject_entry = (GtkWidget*) gtk_object_get_data ( GTK_OBJECT(find_dialog), "subject" );
		g_return_if_fail ( subject_entry!=NULL );
		subject = g_strdup ( gtk_entry_get_text ( GTK_ENTRY(subject_entry) ) );
		g_strstrip ( subject );
		if ( !*subject ) {
			g_free ( subject );
			subject = NULL;
		}

		/* get case sensitivity... */
		case_sensitive_checkbox = (GtkWidget*) gtk_object_get_data ( GTK_OBJECT(find_dialog), "case" );
		g_return_if_fail ( case_sensitive_checkbox!=NULL );
		case_sensitive = gtk_toggle_button_get_active ( GTK_TOGGLE_BUTTON(case_sensitive_checkbox) );

		/* update the find_article fields with these widget values */
		g_free ( find_article__author );
		find_article__author = author;
		g_free ( find_article__subject );
		find_article__subject = subject;
		find_article__case_sensitive = case_sensitive;

		articlelist_find_next_cb ( );
	}
}

void
articlelist_find_next_cb (void)
{
	find_article (
		find_article__author,
		find_article__subject,
		find_article__case_sensitive );
}


void
articlelist_find_text_cb (void)
{
	GtkWidget *table;
	GtkWidget *author_label, *author_entry;
	GtkWidget *subject_label, *subject_entry;
	GtkWidget *case_sensitive_checkbutton;

	/* There can be only one! (find window) */
	if (find_dialog) {
		gdk_window_raise (GTK_WIDGET(find_dialog)->window);
		return;
	}
	
	find_dialog = gnome_dialog_new ( _("Find Message"),
					 GNOME_STOCK_PIXMAP_FORWARD,
					 GNOME_STOCK_PIXMAP_CLOSE, NULL);
	gnome_dialog_set_default ( GNOME_DIALOG(find_dialog), 0 );

	table = gtk_table_new ( 3, 2, FALSE );
//	gtk_container_set_border_width ( GTK_CONTAINER(table), GNOME_PAD_SMALL );
//        gtk_table_set_row_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);
//        gtk_table_set_col_spacings (GTK_TABLE (table), GNOME_PAD_SMALL);

	subject_label = gtk_label_new ( _("Subject:") );
	gtk_table_attach ( GTK_TABLE(table),
			   subject_label,
			   0, 1,
			   0, 1,
			   0,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	subject_entry = gtk_entry_new ( );
	gtk_object_set_data ( GTK_OBJECT(find_dialog), "subject", subject_entry );
	gtk_table_attach ( GTK_TABLE(table),
			   subject_entry,
			   1, 2,
			   0, 1,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	author_label = gtk_label_new ( _("Author:") );
	gtk_table_attach ( GTK_TABLE(table),
			   author_label,
			   0, 1,
			   1, 2,
			   0,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	author_entry = gtk_entry_new ( );
	gtk_object_set_data ( GTK_OBJECT(find_dialog), "author", author_entry );
	gtk_table_attach ( GTK_TABLE(table),
			   author_entry,
			   1, 2,
			   1, 2,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   0,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	case_sensitive_checkbutton = gtk_check_button_new_with_label ( "Case Sensitive" );
	gtk_toggle_button_set_active ( GTK_TOGGLE_BUTTON(case_sensitive_checkbutton), FALSE );
	gtk_object_set_data ( GTK_OBJECT(find_dialog), "case", case_sensitive_checkbutton );
	gtk_table_attach ( GTK_TABLE(table),
			   case_sensitive_checkbutton,
			   0, 2,
			   2, 3,
			   GTK_EXPAND | GTK_FILL | GTK_SHRINK,
			   GTK_EXPAND | GTK_FILL,
			   GNOME_PAD_SMALL, GNOME_PAD_SMALL );

	
//	gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(find_dialog)->vbox), table );
	gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (find_dialog)->vbox), table, FALSE, FALSE, 0);
	
	gtk_signal_connect (GTK_OBJECT(find_dialog), "clicked",
			    GTK_SIGNAL_FUNC(find_dialog_clicked_cb),
			    table);
	gtk_signal_connect (GTK_OBJECT(find_dialog), "destroy",
			    GTK_SIGNAL_FUNC(find_dialog_destroy_cb), NULL);

	gtk_widget_grab_focus (subject_entry);
	gui_popup_draw (find_dialog);
}


/*--------------------------------------------------------------------
 * clear the ctree of all messages
 *--------------------------------------------------------------------*/
void
articlelist_clear (void)
{
	pan_lock();
	gtk_clist_freeze (GTK_CLIST(Pan.article_ctree));
	gtk_clist_clear (GTK_CLIST(Pan.article_ctree));
	gtk_clist_thaw (GTK_CLIST(Pan.article_ctree));
	pan_unlock();
}


/*--------------------------------------------------------------------
 * free an article list (and all threads)
 *--------------------------------------------------------------------*/
void
articlelist_free (GSList *alist)
{
	GSList *p;
	article_data *adata;

	for (p = alist; p != NULL; p = p->next) {
/*		g_assert (p->data); */
		adata = p->data;

		if (adata->threads)
			articlelist_free (adata->threads);
		article_free (adata);
		p->data = NULL;
	}

	g_slist_free (alist);
}


void
articlelist_get_all (void)
{
	if (Pan.active_group)
		queue_add_articlelist_download (Pan.active_group,
						QUEUE_ARTICLELIST_DOWNLOAD_ALL |
						QUEUE_FOREGROUND);
}


void
articlelist_get_new (void)
{
	if (Pan.active_group)
		queue_add_articlelist_download (Pan.active_group,
						QUEUE_ARTICLELIST_DOWNLOAD_NEW |
						QUEUE_FOREGROUND);
}


/*--------------------------------------------------------------------
 * load article listing
 *--------------------------------------------------------------------*/
void
articlelist_load (queue_data *qdata)
{
	group_data *gdata = NULL;
	GSList *alist;

	/* Clear the currently displayed articlelist, if any */
	if (Pan.active_group) {
		articlelist_clear ();
		articlelist_free (Pan.active_group->articlelist);
		Pan.active_group->articlelist = NULL;
	}

	gdata = qdata->gdata;

	/* step 1: load article headers */
	if ((alist = db_load_articles (gdata)) == NULL) {
		pan_lock ();
		choose_download_dialog (gdata);
		pan_unlock ();
	} 
	else {
		grouplist_update_row (gdata);
//		gui_set_info_va (_("Loaded %d article headers"), group_get_attrib_i (gdata, "Total"));
		log_add_va (_("Loaded %d article headers for group %s"), group_get_attrib_i (gdata, "Total"), gdata->name);

		/* step 2: thread article list */
		gui_set_status (_("Threading article list"));
                alist = thread_articles ( alist, gdata );
		gdata->articlelist = alist;

		/* step 3: display article list */
		gui_set_status (_("Displaying article list"));
		articlelist_update ((group_data *) gdata);

//		gui_clean ();
	}
}


/*--------------------------------------------------------------------
 * this function removes the STATE_NEW flag from every article in 
 * the list, called when 'get new articles' is run, to mark all as old
 *--------------------------------------------------------------------*/
void
articlelist_not_new (GSList *l)
{
	for (; l; l = l->next) {
		((article_data *)l->data)->state &= ~STATE_NEW;
		if (((article_data *)l->data)->threads) {
			articlelist_not_new (((article_data *)l->data)->threads);
		}
	}
}

/*--------------------------------------------------------------------
 * process the selected articles
 *--------------------------------------------------------------------*/
static
void
articlelist_process_selected (int type)
{
	GList *p;
	article_data *adata;

	for (p = GTK_CLIST(Pan.article_ctree)->selection; p; p = p->next) {
		adata = (article_data *) gtk_ctree_node_get_row_data(GTK_CTREE(Pan.article_ctree), p->data);
		if (adata) {
			switch (type) {
			case 2:
				queue_add_article_decode(adata, QUEUE_ARTICLE_DECODE | QUEUE_BACKGROUND);
				adata->state |= STATE_DECODE_QUEUED;
				/* Calling article_update here is too slow */
				articlelist_update_node (adata);
				break;
			case 3: /* mark read */
				article_mark_read (adata);
				articlelist_update_node (adata);
				break;
			case 4: /* mark unread */
				article_mark_unread (adata);
				articlelist_update_node (adata);
				break;
			case 5: /* delete */
				article_delete ();
				break;
			case 6:
				queue_add_article_decode(adata, QUEUE_ARTICLE_DECODE | QUEUE_BACKGROUND | QUEUE_OPEN);
				adata->state |= STATE_DECODE_QUEUED;
				/* Calling article_update here is too slow */
				articlelist_update_node (adata);
				break;
			default:
				break;
			}
		}
	}
	gtk_ctree_unselect_recursive (GTK_CTREE(Pan.article_ctree), NULL);
}

void
articlelist_selected_download (void)
{
	articlelist_process_selected (0);
}

void
articlelist_selected_decode (void)
{
	articlelist_process_selected (2);
}

void
articlelist_selected_mark_read (void)
{
	articlelist_process_selected (3);
}

void articlelist_selected_mark_unread (void)
{
	articlelist_process_selected (4);
}

void
articlelist_selected_delete (void)
{
	articlelist_process_selected (5);
}

void
articlelist_selected_open (void)
{
	articlelist_process_selected (6);
}

void
article_ctree_clean (void)
{
	gtk_clist_clear (GTK_CLIST (Pan.article_ctree));
	if (Pan.active_group) {
		group_clean (Pan.active_group);
		Pan.active_group = NULL;
                find_article_next_pos = NULL;
	}
}


static GnomePixmap*
get_article_pixmap ( const article_data* adata )
{
	g_return_val_if_fail ( adata!=NULL, NULL );

	if (adata->state & STATE_ERROR)
		return error_pixmap;

	if (adata->state & STATE_DECODED)
		return binary_decoded_pixmap;

	if (adata->state & STATE_READ)
		return article_read_pixmap;

	if (adata->state & STATE_DOWNLOAD_QUEUED)
		return download_queue_pixmap;

	if (adata->state & STATE_DECODE_QUEUED)
		return decode_queue_pixmap;

	if (adata->state & STATE_MULTIPART_ALL)
		return mp_complete_pixmap;

	if (adata->state & STATE_MULTIPART_SOME)
		return mp_incomplete_pixmap;

	return NULL;
}

inline void
articlelist_select_all (void)
{
	gtk_ctree_select_recursive (GTK_CTREE(Pan.article_ctree), NULL);
}


static void
articlelist_init_styles (void)
{
	unread_style = gtk_style_copy (gtk_widget_get_style (GTK_WIDGET (Pan.window)));
	read_style = gtk_style_copy (gtk_widget_get_style (GTK_WIDGET (Pan.window)));
	new_style = gtk_style_copy (gtk_widget_get_style (GTK_WIDGET (Pan.window)));

	unread_style->font = gdk_font_load (articlelist_font);
	read_style->font = gdk_font_load (articlelist_font);
	new_style->font = gdk_font_load (articlelist_font);

	memcpy (unread_style->fg, &unread_color, sizeof (GdkColor));
	memcpy (read_style->fg, &read_color, sizeof (GdkColor));
	memcpy (new_style->fg, &new_color, sizeof (GdkColor));
}


GtkCTreeNode*
add_article_to_ctree (GtkCTreeNode *parent, article_data *adata)
{
	char *text[4];
	char line_count_buf[16];
	GtkCTreeNode *node;
	GnomePixmap* icon = get_article_pixmap ( adata );

        /* this is sortable with a simple strcmp */
        char timebuf[32];
        const struct tm local_tm = *localtime ( &adata->date );
        strftime ( timebuf, sizeof(timebuf), "%Y/%m/%d %H:%M:%S", &local_tm );

	if ( articlelist_font && !unread_style )
		articlelist_init_styles ();

	/* line count */
	if ( ! ( adata->state & STATE_MULTIPART_ALL ) ) {
		sprintf ( line_count_buf, "%d", adata->linecount );
	}
	else  { /* multipart messages give total line count... */
		guint lcountsum = adata->linecount;
		GSList *l;
		for ( l = adata->threads; l; l = l->next)
			lcountsum += ((article_data *)l->data)->linecount;
		sprintf ( line_count_buf, "%d", lcountsum );
	}

	text[0] = adata->subject;
	text[1] = line_count_buf;
	text[2] = timebuf;
	text[3] = adata->author;

	node = gtk_ctree_insert_node (
		GTK_CTREE(Pan.article_ctree),
		parent, NULL,
		text, 5,
		(icon ? icon->pixmap : NULL),
		(icon ? icon->mask : NULL),
		(icon ? icon->pixmap : NULL),
		(icon ? icon->mask : NULL),
		FALSE, FALSE );


/* Set text color and font */
	if ( articlelist_font ) {
		if ((adata->state & STATE_READ) || (adata->state & STATE_DECODED))
			gtk_ctree_node_set_row_style (GTK_CTREE(Pan.article_ctree), node, read_style);
		else if (adata->state & STATE_NEW)
			gtk_ctree_node_set_row_style (GTK_CTREE(Pan.article_ctree), node, new_style);
		else
			gtk_ctree_node_set_row_style (GTK_CTREE(Pan.article_ctree), node, unread_style);
	}	
/* If the font isn't specified by the user, just set the color */
	else {
		if ((adata->state & STATE_READ) || (adata->state & STATE_DECODED))
			gtk_ctree_node_set_foreground (GTK_CTREE(Pan.article_ctree), node, &read_color);
		else if (adata->state & STATE_NEW)
			gtk_ctree_node_set_foreground (GTK_CTREE(Pan.article_ctree), node, &new_color);
		else
			gtk_ctree_node_set_foreground (GTK_CTREE(Pan.article_ctree), node, &unread_color);
	}

	gtk_ctree_node_set_row_data (GTK_CTREE(Pan.article_ctree), node, (article_data *) adata);

	adata->node = node;

	return node;
}


int
build_article_ctree_recursive ( GtkCTreeNode* parent,
				GSList* list )
{
	int count = 0;

	GSList *pos;
        for ( pos=list; pos!=NULL; pos=pos->next )
        {
		article_data* data = (article_data *) pos->data;
		GtkCTreeNode* sibling = add_article_to_ctree (parent, data);
		++count;
		if ( data->nthreads ) //&& !( data->state & STATE_MULTIPART_ALL ) )
			count += build_article_ctree_recursive (sibling, data->threads);
        }

        return count;
}


static void
articlelist_sort_init ( group_data *gdata)
{
	gint sort_style = 0;

	sort_style = group_get_attrib_i (gdata, "sort_style");

	if ( sort_style > 3 )
		GTK_CLIST (Pan.article_ctree)->sort_type = GTK_SORT_DESCENDING;
	else
		GTK_CLIST (Pan.article_ctree)->sort_type = GTK_SORT_ASCENDING;

        gtk_clist_set_sort_column ( GTK_CLIST(Pan.article_ctree), sort_style % 4 );
}


void
articlelist_update ( group_data *gdata )
{
	char *text[4];
	GtkCTreeNode *parent;

	if (gdata == NULL)
		return;

	text[0] = gdata->name;
	text[1] = text[2] = text[3] = NULL;

	gui_set_status (_("Updating article display"));
	articlelist_clear ();

	pan_lock ();

	gtk_clist_freeze (GTK_CLIST (Pan.article_ctree));
	parent = gtk_ctree_insert_node (GTK_CTREE(Pan.article_ctree), NULL, NULL,
					text, 5, articlelist_closed_pixmap->pixmap,
					articlelist_closed_pixmap->mask,
					articlelist_open_pixmap->pixmap,
					articlelist_open_pixmap->mask, FALSE, TRUE);

	build_article_ctree_recursive (parent, gdata->articlelist);

	/* sort the newly updated articlelist*/
	articlelist_sort_init ( gdata );
	articlelist_sort ( );

//	if (Pan.active_group && (Pan.active_group != gdata))
//		group_clean (Pan.active_group);

	Pan.active_group = gdata;
	find_article_next_pos = NULL;

	gtk_clist_thaw (GTK_CLIST (Pan.article_ctree));
	pan_unlock ();
}


void 
articlelist_update_node (article_data *adata)
{
	GtkCTreeNode *node = NULL;

	if ( articlelist_font && !unread_style )
		articlelist_init_styles ();

	g_assert (adata && adata->message_id);

	if (adata->node)
		node = adata->node;
	else
		node = gtk_ctree_find_by_row_data (GTK_CTREE(Pan.article_ctree), NULL, adata);


	if (node)
	{
		/* set icon */
		GnomePixmap* icon = get_article_pixmap ( adata );
		gtk_ctree_set_node_info (
			GTK_CTREE(Pan.article_ctree), node,
			adata->subject, 5,
			(icon ? icon->pixmap : NULL),
			(icon ? icon->mask : NULL),
			(icon ? icon->pixmap : NULL),
			(icon ? icon->mask : NULL),
			GTK_CTREE_ROW(node)->is_leaf,
			GTK_CTREE_ROW(node)->expanded );

/* Set text color and font */
		if ( articlelist_font  ) {
			if ((adata->state & STATE_READ) || (adata->state & STATE_DECODED))
				gtk_ctree_node_set_row_style (GTK_CTREE(Pan.article_ctree), node, read_style);
			else if (adata->state & STATE_NEW)
				gtk_ctree_node_set_row_style (GTK_CTREE(Pan.article_ctree), node, new_style);
			else
				gtk_ctree_node_set_row_style (GTK_CTREE(Pan.article_ctree), node, unread_style);
		}	
/* If the font isn't specified by the user, just set the color */
		else {
			if ((adata->state & STATE_READ) || (adata->state & STATE_DECODED))
				gtk_ctree_node_set_foreground (GTK_CTREE(Pan.article_ctree), node, &read_color);
			else if (adata->state & STATE_NEW)
				gtk_ctree_node_set_foreground (GTK_CTREE(Pan.article_ctree), node, &new_color);
			else
				gtk_ctree_node_set_foreground (GTK_CTREE(Pan.article_ctree), node, &unread_color);
		}
		
	}
}


/*********************************************************************
 * The following stuff implements the "thread articles" functionality
 ********************************************************************/

static   
int 
mp_strcasecmp (
	register const char* mp,
	register const char* ph )
{
	while ( *mp && *ph )
		if ((*mp=='('&&*ph=='(') || (*mp=='['&&*ph=='[')) {
			++mp;
			++ph;
			if ( isdigit(*mp) && isdigit(*ph) ) {
				while ( isdigit(*mp) ) ++mp;
				while ( isdigit(*ph) ) ++ph;
			}
		}
		else if ( toupper(*mp) == toupper(*ph) ) {
			++mp;
			++ph;
		}
		else return toupper(*mp) - toupper(*ph);

	return 0;
}


/**
 * Skip the "Re: " part of a subject header, if any
 * @param subject
 * @return the non-"Re:" portion of the subject header
 */
static
const char*
skip_reply_leader ( const char* a )
{
	if ( !a )
		return NULL;

	if ( toupper(a[0])=='R' && toupper(a[1])=='E' && toupper(a[2])==':' )
		return a+4;

	return a;
}

/**
 * Sort by message id
 */
static
int
gslist_message_id_comp (
	const void* va,
	const void* vb )
{
	const article_data* a = ((const article_data*)((*(const GSList**)va)->data));
	const article_data* b = ((const article_data*)((*(const GSList**)vb)->data));
	return strcmp ( a->message_id, b->message_id );
}

/**
 * Works like article_subject_cmp, but handles multiparts better
 */
static
int
article_mp_subject_comp (
	const void* va,
	const void* vb )
{
	const article_data* a = (const article_data*)va;
	const char* subj_a = skip_reply_leader ( a->subject );
	const gboolean a_is_re = a->subject != subj_a;

	const article_data* b = (const article_data*)vb;
	const char* subj_b = skip_reply_leader ( b->subject );
	const gboolean b_is_re = b->subject != subj_b;

	/* get the easy stuff out of the way */
	int value = mp_strcasecmp ( subj_a, subj_b );
	if ( value )
		return value;

	/* check multipart */
	value = a->part - b->part;
	if ( value )
		return value;

	/* if one but not both is a reply, the reply goes second */
	if ( !!a_is_re != !!b_is_re )
		return a_is_re ? 1 : -1;

	/* oldest goes first... */
	value = a->date - b->date;
	if ( value )
		return value;

	return 0;
}

/**
 * Sort thusly:
 * (1) by subject, with replies coming after non-replies of the same subject, but before other subjects.
 * (2) by time
 */
static
int
article_subject_comp (
	const void* va,
	const void* vb )
{
	const article_data* a = (const article_data*)va;
	const char* subj_a = skip_reply_leader ( a->subject );
	const gboolean a_is_re = a->subject != subj_a;

	const article_data* b = (const article_data*)vb;
	const char* subj_b = skip_reply_leader ( b->subject );
	const gboolean b_is_re = b->subject != subj_b;

	/* if they're not related, a simple strcmp does the job */
	int value = g_strcasecmp ( subj_a, subj_b );
	if ( value )
		return value;

	/* if one but not both is a reply, the reply goes second */
	if ( !!a_is_re != !!b_is_re )
		return a_is_re ? 1 : -1;

	/* oldest goes first... */
	value = a->date - b->date;
	if ( value )
		return value;

	return 0;
}

/**
 * Wrapper around article_subject_comp to compare to GSList nodes.
 */
static
int
gslist_subject_comp (
	const void* va,
	const void* vb )
{
	return article_subject_comp (
		((const article_data*)((*(const GSList**)va)->data)),
		((const article_data*)((*(const GSList**)vb)->data)) );
}

/**
 * Sort thusly:
 * (1) in ascending number of part qty
 * (2) by subject via mp_strcasecmp
 * (3) by part number
 * This should lump all the multiparts together in ascending order.
 */
static
int
gslist_multipart_comp (
	const void* va,
	const void* vb )
{
	const article_data* a = (const article_data*)((*(const GSList**)va)->data);
	const article_data* b = (const article_data*)((*(const GSList**)vb)->data);

	/* parts */
	int val = a->parts - b->parts;
	if ( val )
		return val;

	/* subject */
	val = *a->subject - *b->subject;
	if ( val )
		return val;
	val = mp_strcasecmp ( a->subject, b->subject );
	if ( val )
		return val;

	/* part number */
	val = a->part - b->part;
	if ( val )
		return val;

	/* did someone repost this? */
	val = a->date - b->date;
	if ( val )
		return val;

	/* they are equal */
	return 0;
}

/**
 * Thread the articles specified in list
 */
GSList*
thread_articles (
	GSList* list,
	group_data* gdata )
{
	int i, j;
	GSList **buf, **refs, **subjs, **parts, *p;
	GSList search_gsl, *search_gsl_p;
	article_data search_a;
	int remove_qty = 0;
	int* remove_from_top = NULL;
	int article_qty;

	/* if nothing to do, do nothing */
	article_qty = g_slist_length ( list );
	if ( article_qty != group_get_attrib_i ( gdata, "Total" ) ) 
		g_warning ( "group_get_attrib_i(gdata,\"Total\") is in error" );
	if ( !article_qty )
		return list;

	/* make a reference copy of the articles... */
	buf = g_new ( GSList*, article_qty );
	for ( i=0, p=list; p!=NULL; p=p->next, ++i )
		buf[i] = p;

	/* make various sorted copies of the articles for fast searching */
	refs = g_memdup ( buf, sizeof(GSList*)*article_qty );
	subjs = g_memdup ( buf, sizeof(GSList*)*article_qty );
	parts = g_memdup ( buf, sizeof(GSList*)*article_qty );
	qsort ( refs, article_qty, sizeof(GSList*), gslist_message_id_comp );
	qsort ( subjs, article_qty, sizeof(GSList*), gslist_subject_comp );
	qsort ( parts, article_qty, sizeof(GSList*), gslist_multipart_comp );

	/* housekeeping */
	search_gsl_p = &search_gsl;
	search_gsl.data = &search_a;
	search_gsl.next = NULL;
	remove_qty = 0;
	remove_from_top = g_new ( int, article_qty );
	gui_progress_init_steps ( article_qty );

	/* thread the articles */
	for ( i=0; i!=article_qty; ++i )
	{
		article_data* a = (article_data*)(buf[i]->data);
		gboolean exact_match = FALSE;
		int index = -1;
		p = NULL;

		/* visual feedback */
		gui_progress_next_step ( );

		/* thread by reference
		   (except for multiparts, which should be top-level -- so add !a->parts ) */
		if ( a->references && *a->references=='<' && !a->parts )
		{
			search_a.message_id = strrchr ( a->references, '<' );
			index = lower_bound ( &search_gsl_p, refs, article_qty, sizeof(GSList*), gslist_message_id_comp, &exact_match );

			if ( exact_match )
				p = refs[index];
		}

		/* thread by multipart */
		if ( p==NULL && a->parts>1 && a->part>1 )
		{
			search_a.subject = a->subject;
			search_a.part = 1;
			search_a.parts = a->parts;
			search_a.date = 0; /* damned unlikely to get an exact match now. :) */
			index = lower_bound ( &search_gsl_p, parts, article_qty, sizeof(GSList*), gslist_multipart_comp, &exact_match );

			if ( index<article_qty && parts[index]!=buf[i] )
				p = parts[index];
		}

		/* thread by subject */
		if ( p==NULL && skip_reply_leader(a->subject)!=a->subject )
		{
			search_a.subject = (char*) skip_reply_leader ( a->subject );
			search_a.date = 0; /* damned unlikely to get an exact match now. :) */
			index = lower_bound ( &search_gsl_p, subjs, article_qty, sizeof(GSList*), gslist_subject_comp, &exact_match );

			if ( index<article_qty ) {
				article_data* b = (article_data*)(subjs[index]->data);
				if ( !g_strcasecmp ( search_a.subject, b->subject ) ) /* simple case of source/reply */
					p = subjs[index];
				else if ( !g_strcasecmp ( a->subject, b->subject ) && b->date<a->date ) /* 2 replies, no source: thread under oldest */
					p = subjs[index];
			}
		}

		/* this article has a parent */
		if ( p!=NULL )
		{
			article_data* parent = (article_data*)(p->data);
			parent->threads = g_slist_insert_sorted ( parent->threads, a, article_mp_subject_comp );
			++parent->nthreads;
			remove_from_top[remove_qty++] = i;
		}
	}

	gui_set_progress ( 100.0 );

	g_free ( parts );
	g_free ( subjs );

	/* NULL out the elements to be removed from top level */
	for ( i=0; i!=remove_qty; ++i ) {
		g_slist_free_1 ( buf[remove_from_top[i]] );
		buf[remove_from_top[i]] = NULL;
	}
	g_free ( remove_from_top );

	/* remove the NULL elements.
	   We recycle refs here for convenience */
	for ( i=j=0; i!=article_qty; ++i )
		if ( buf[i]!=NULL )
			refs[j++] = buf[i];
	article_qty = j;
	g_free ( buf );
	buf = refs;
	refs = NULL;

	/* set the multipart state (all/partial) */
	for ( i=0; i!=article_qty; ++i ) {
		article_data* a = (article_data*)(buf[i]->data);
		if ( a->parts<2 )
			continue;
		if ( a->part==0 ) /* this is treated separately, so don't give it a multipart state */
			continue;
		if ( a->part > 1 ) { /* if a top-level's part is >1, we already know it's incomplete */
			a->state |= STATE_MULTIPART_SOME;
			continue;
		}
		/* make sure we have each multipart. */
		for ( j=a->part+1, p=a->threads; j!=a->parts && p!=NULL; p=p->next )
		{
			article_data* b = (article_data*)(p->data);
			if ( b->part > j )
				break;
			else if ( b->part == j )
				++j;
			else
				; /* a repost of a multipart section? */
		}
		if ( j==a->parts ) {
			a->state |= STATE_MULTIPART_ALL;
		} else {
			a->state |= STATE_MULTIPART_SOME;
		}
	}

	/* finalize the top-level list */
	for ( i=0, j=article_qty-1; i!=j; ++i )
		buf[i]->next = buf[i+1];
	buf[j]->next = NULL;
	list = buf[0];
	g_free ( buf );

	return list;
}


/***************************************************************
 * The following stuff implements the tree sorting functionality
 **************************************************************/

void
articlelist_sort ( void )
{
        const int sort_column = GTK_CLIST(Pan.article_ctree)->sort_column;

        switch ( sort_column )
        {
	case 0:
		gtk_clist_set_compare_func ( GTK_CLIST (Pan.article_ctree), (GtkCListCompareFunc) articlelist_compare_subjects );
		break;

	case 1:
		gtk_clist_set_compare_func (GTK_CLIST (Pan.article_ctree), (GtkCListCompareFunc) articlelist_compare_lines );
		break;

	case 4:
		gtk_clist_set_compare_func (GTK_CLIST (Pan.article_ctree), (GtkCListCompareFunc) articlelist_compare_text );
		break;

	default:
		gtk_clist_set_compare_func (GTK_CLIST (Pan.article_ctree), NULL);
		break;
        }

	gtk_clist_sort (GTK_CLIST (Pan.article_ctree));
}


static
gint
articlelist_compare_lines (GtkCList *clist,
			   gconstpointer ptr1,
			   gconstpointer ptr2)
{
	const char* text1 = GTK_CELL_TEXT(((GtkCListRow*)(ptr1))->cell[clist->sort_column])->text;
	const char* text2 = GTK_CELL_TEXT(((GtkCListRow*)(ptr2))->cell[clist->sort_column])->text;

	if (!text2)
		return (text1 != NULL);

	if (!text1)
		return -1;

        return atoi(text1) - atoi(text2);
}

static
gint
articlelist_compare_subjects (
        GtkCList* clist,
        gconstpointer ptr1,
        gconstpointer ptr2 )
{
	const GtkCListRow *row1 = (const GtkCListRow *) ptr1;
	const char *text1 = GTK_CELL_TEXT ( row1->cell[clist->sort_column] )->text;

	const GtkCListRow *row2 = (const GtkCListRow *) ptr2;
	const char *text2 = GTK_CELL_TEXT ( row2->cell[clist->sort_column] )->text;

	if (!text2)
		return (text1 != NULL);

	if (!text1)
		return -1;

        text1 = skip_reply_leader ( text1 );
        text2 = skip_reply_leader ( text2 );
        return g_strcasecmp ( text1, text2 );
}


static
gint
articlelist_compare_text (GtkCList *clist,
			  gconstpointer ptr1,
			  gconstpointer ptr2)
{
	const GtkCListRow *row1 = (const GtkCListRow *) ptr1;
	const char *text1 = GTK_CELL_TEXT ( row1->cell[clist->sort_column] )->text;

	const GtkCListRow *row2 = (const GtkCListRow *) ptr2;
	const char *text2 = GTK_CELL_TEXT ( row2->cell[clist->sort_column] )->text;

	if (!text2)
		return (text1 != NULL);

	if (!text1)
		return -1;

	return g_strcasecmp ( text1, text2 );
}


static
void
articlelist_click_column (GtkCList *clist, int n)
{
	gint sort_style;

	if (clist->sort_column == n) { /* flip sort styles if the column was clicked twice */
		if ( clist->sort_type == GTK_SORT_ASCENDING ) {
			clist->sort_type = GTK_SORT_DESCENDING;
		}
		else {
			clist->sort_type =  GTK_SORT_ASCENDING;
		}
	}
	else
		clist->sort_type = GTK_SORT_ASCENDING;

	gtk_clist_set_sort_column (GTK_CLIST (clist), n);

        articlelist_sort ( );

	sort_style = n + (clist->sort_type == GTK_SORT_DESCENDING)*4;

	group_set_attrib_i (Pan.active_group, "sort_style", sort_style);
}


static void article_ctree_destroy_cb ( void )
{
	gtk_widget_destroy ( GTK_WIDGET ( articlelist_closed_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( articlelist_open_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( article_read_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( decode_queue_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( download_queue_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( error_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( binary_decoded_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( mp_complete_pixmap ) );
	gtk_widget_destroy ( GTK_WIDGET ( mp_incomplete_pixmap ) );
}


/*--------------------------------------------------------------------
 * generate the listing of articles, for the "Articles" tab
 *--------------------------------------------------------------------*/
GtkWidget*
create_articlelist_ctree (void)
{
	GtkWidget *s_window;

	char *titles[] = {
		_("Subject"),
		_("Lines"),
		_("Date"),
		_("Author"),
	};

	articlelist_closed_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(book_closed_xpm);
	articlelist_open_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(book_open_xpm);
	article_read_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(mini_page_xpm);
	decode_queue_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(bluecheck_xpm);
	download_queue_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(greencheck_xpm);
	error_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(x_xpm);
	binary_decoded_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(binary_xpm);
	mp_complete_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(binary_complete_xpm);
	mp_incomplete_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d(binary_incomplete_xpm);

	cmap = gdk_colormap_get_system ();
	if (!gdk_color_alloc (cmap, &unread_color))
		g_error ("couldn't allocate color");
	if (!gdk_color_alloc (cmap, &new_color))
		g_error ("couldn't allocate color");
	if (!gdk_color_alloc (cmap, &read_color))
		g_error ("couldn't allocate color");

	Pan.article_ctree = gtk_ctree_new_with_titles (4, 0, titles);
	gtk_ctree_set_line_style (GTK_CTREE(Pan.article_ctree),
				  GTK_CTREE_LINES_DOTTED);
	gtk_widget_show (Pan.article_ctree);

	s_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW(s_window),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(s_window), Pan.article_ctree);

	gtk_clist_set_selection_mode (GTK_CLIST(Pan.article_ctree),
				      GTK_SELECTION_EXTENDED);

	gtk_clist_set_column_width (GTK_CLIST(Pan.article_ctree), 0, 400);
	gtk_clist_set_column_auto_resize (GTK_CLIST(Pan.article_ctree), 1, TRUE);
	gtk_clist_set_column_auto_resize (GTK_CLIST(Pan.article_ctree), 2, TRUE);

	gtk_clist_set_column_justification (GTK_CLIST (Pan.article_ctree), 2,
					    GTK_JUSTIFY_RIGHT);

/* create the right click popup menu */
	article_ctree_menu = gnome_popup_menu_new (articlelist_menu_popup);

/* connect signals */
	gtk_signal_connect (GTK_OBJECT(Pan.article_ctree), "button_press_event",
			    GTK_SIGNAL_FUNC(articlelist_button_press), NULL);

	gtk_signal_connect (GTK_OBJECT(Pan.article_ctree), "click_column",
			    GTK_SIGNAL_FUNC(articlelist_click_column), NULL);

	gtk_signal_connect (GTK_OBJECT(Pan.article_ctree), "destroy",
	                    GTK_SIGNAL_FUNC(article_ctree_destroy_cb), NULL);


	return s_window;
}
